/*
 * Copyright 2008,2009 WSO2, Inc. http://www.wso2.org.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.mercury.state;

import org.wso2.mercury.message.CreateSequenceMessage;
import org.wso2.mercury.exception.RMMessageBuildingException;
import org.wso2.mercury.util.RMUtil;
import org.wso2.mercury.util.MercuryConstants;
import org.wso2.mercury.workers.RMDSequenceWorker;
import org.wso2.mercury.workers.InvokerWorker;
import org.wso2.mercury.workers.RMSSequenceWorker;
import org.wso2.mercury.keys.InternalSequenceKey;
import org.wso2.mercury.persistence.exception.PersistenceException;
import org.wso2.mercury.persistence.dto.RMDSequenceDto;
import org.wso2.mercury.persistence.PersistenceManager;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.addressing.AddressingConstants;
import org.apache.axis2.description.WSDL2Constants;
import org.apache.axis2.description.ClientUtils;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.AxisFault;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.util.*;

/**
 * this class is used to keep the RMD context
 * objects.
 */
public class RMDContext {

    private static Log log = LogFactory.getLog(RMDContext.class);
    // keep the RMD with the sequnce ID
    private Map sequenceIDRMDSequenceMap;

    private ConfigurationContext configurationContext;

    public RMDContext(ConfigurationContext configurationContext) {
        this.sequenceIDRMDSequenceMap = new HashMap();
        this.configurationContext = configurationContext;
    }

    /**
     * creates an RMD sequence at the RMD side once it gets a create sequence message
     * this sequence can eiter be an annonymous or duplex one. for annonymous we let the message to
     * countine and write the response at the out handler
     * @param messageContext
     */
    public synchronized RMDSequence processCreateSequenceRequest(MessageContext messageContext)
            throws RMMessageBuildingException, AxisFault {
        // first check whether there is a sequence alrady started for this
        // acks to
        CreateSequenceMessage createSequenceMessage =
                CreateSequenceMessage.fromSOAPEnvelope(messageContext.getEnvelope());

        // otherwise this may be a duplicate create seqeunce message
        String sequenceID = RMUtil.getUUID();

        // creates a sequence object for this sequence
        RMDSequence rmdSequence = new RMDSequence(RMDSequence.STATE_00);
        rmdSequence.setAcksTo(new EndpointReference(createSequenceMessage.getAcksToAddress()));
        rmdSequence.getInvokerBuffer().setAcksTo(new EndpointReference(createSequenceMessage.getAcksToAddress()));
        rmdSequence.setSequenceID(sequenceID);
        rmdSequence.setCreateSequenceMessageContext(messageContext);

        Axis2Info axis2Info = getAxis2Info(messageContext,
                createSequenceMessage.getAcksToAddress(),
                createSequenceMessage.getAddressingNamespace());
        rmdSequence.setAxis2Info(axis2Info);
        rmdSequence.setInvokerBufferPersistanceManager();

        if (messageContext.getConfigurationContext().getProperty(MercuryConstants.RM_PERSISTANCE_MANAGER) != null){
            // i.e RMDSequence must be persisted.
            try {
                rmdSequence.save();
            } catch (PersistenceException e) {
                log.error("Can not save to data base");
            }

        }


        //register this sequence in the Application context
        registerRMDSequenceToSequenceID(sequenceID, rmdSequence);
        String acksToAddress = createSequenceMessage.getAcksToAddress();
        // process the offer ID
        if (createSequenceMessage.getOfferIdentifier() != null) {
            if (messageContext.getAxisOperation().getMessageExchangePattern().equals(WSDL2Constants.MEP_URI_IN_OUT)) {

                // set the self acts to EPR in RMDSequence
                rmdSequence.setSelfAcksToEPR(messageContext.getOptions().getTo().getAddress());

                RMSSequence rmsSequence =
                        new RMSSequence(RMSSequence.STATE_1000,
                                new EndpointReference(createSequenceMessage.getAcksToAddress()));
                // setting the wso2 info
                axis2Info = getAxis2Info(messageContext,
                        createSequenceMessage.getAcksToAddress(),
                        createSequenceMessage.getAddressingNamespace());
                // this can only happen at the server side and there to means this server address
                // so we have to change the to to the RMD address.
                axis2Info.getOptions().setTo(new EndpointReference(createSequenceMessage.getAcksToAddress()));
                rmsSequence.setAxis2Info(axis2Info);
                rmsSequence.setSequenceID(createSequenceMessage.getOfferIdentifier());

                // no need to set the acks to epr since no need to send the create sequence message
                // register this sequence in the application context
                InternalSequenceKey key = new InternalSequenceKey(createSequenceMessage.getAcksToAddress(), sequenceID);
                RMSContext rmsContext = (RMSContext) this.configurationContext.getProperty(MercuryConstants.RMS_CONTEXT);
                rmsContext.registerRMSSequenceToInternalKey(key, rmsSequence);
                rmsContext.registerRMSSequenceToSequenceID(createSequenceMessage.getOfferIdentifier(),rmsSequence);

                if (acksToAddress.equals(AddressingConstants.Submission.WSA_ANONYMOUS_URL) ||
                        acksToAddress.equals(AddressingConstants.Final.WSA_ANONYMOUS_URL)) {
                    rmsSequence.setAnnonymous(true);
                    rmdSequence.setOfferedRMSSequence(rmsSequence);
                    rmsSequence.setOfferedRMDSequence(rmdSequence);
                    // we don't have to start any threads here since this is in the server side
                } else {
                    rmsSequence.setAnnonymous(true);
                     // set a sender worker who works on this sequence.
                    this.configurationContext.getThreadPool().execute(new RMSSequenceWorker(rmsSequence));
                }
            }
        }

        // if the create sequence message has an anonymous acks to address
        // then we can't initiate the threads from here.

        // start the InvokerWorker to pass the message to next level
        InvokerWorker invokerWorker = new InvokerWorker(rmdSequence.getInvokerBuffer());
        configurationContext.getThreadPool().execute(invokerWorker);


        if (acksToAddress.equals(AddressingConstants.Submission.WSA_ANONYMOUS_URL) ||
                acksToAddress.equals(AddressingConstants.Final.WSA_ANONYMOUS_URL)) {
            rmdSequence.setAnonymous(true);
            rmdSequence.getInvokerBuffer().setAnonymous(true);
        } else {
            rmdSequence.setAnonymous(false);
            rmdSequence.getInvokerBuffer().setAnonymous(false);
            // start the RMDSequenceWorker to poll the seqence.
            RMDSequenceWorker rmdSequenceWorker = new RMDSequenceWorker(rmdSequence);
            configurationContext.getThreadPool().execute(rmdSequenceWorker);
        }
        return rmdSequence;
    }

    public synchronized void loadRMDSequenceFromPersistanceStorage(String sequenceID,
                                                                   MessageContext messageContext)
            throws PersistenceException, AxisFault {
        PersistenceManager persistenceManager =
                (PersistenceManager) messageContext.getConfigurationContext().getProperty(
                        MercuryConstants.RM_PERSISTANCE_MANAGER);
        RMDSequenceDto rmdSequenceDto = persistenceManager.getRMDSequeceWithSequenceID(sequenceID);
        RMDSequence rmdSequence = new RMDSequence(rmdSequenceDto.getState());
        rmdSequence.setSequenceID(rmdSequenceDto.getSequenceID());
        rmdSequence.setAcksTo(new EndpointReference(rmdSequenceDto.getAcksTo()));
        rmdSequence.setLastMessageNumber(rmdSequenceDto.getLastMessageNumber());
        rmdSequence.setPersistanceDto(rmdSequenceDto);

        //creating the Axis2Info
        //TODO: set this correctly
        Axis2Info axis2Info = getAxis2Info(messageContext,rmdSequenceDto.getAcksTo(), null);
        rmdSequence.setAxis2Info(axis2Info);
        rmdSequence.setInvokerBufferPersistanceManager();
        rmdSequence.loadRMDSequenceDetails(messageContext);

        //register this sequence in the Application context
        registerRMDSequenceToSequenceID(rmdSequence.getSequenceID(), rmdSequence);

        // start the RMDSequenceWorker to poll the seqence.
        RMDSequenceWorker rmdSequenceWorker = new RMDSequenceWorker(rmdSequence);
        configurationContext.getThreadPool().execute(rmdSequenceWorker);

        // start the InvokerWorker to pass the message to next level
        InvokerWorker invokerWorker = new InvokerWorker(rmdSequence.getInvokerBuffer());
        configurationContext.getThreadPool().execute(invokerWorker);


    }

    private Axis2Info getAxis2Info(MessageContext messageContext,
                                   String acksToAddress,
                                   String addressingNamespaceURI) throws AxisFault {
        Axis2Info axis2Info = new Axis2Info();
        axis2Info.setServiceContext(messageContext.getServiceContext());
        axis2Info.setOptions(messageContext.getOptions());
        axis2Info.setTransportIn(messageContext.getTransportIn());

        TransportOutDescription transportOut = ClientUtils.inferOutTransport(
                messageContext.getConfigurationContext().getAxisConfiguration(),
                new EndpointReference(acksToAddress), messageContext);

        axis2Info.setTransportOut(transportOut);
        axis2Info.setServerSide(messageContext.isServerSide());
        axis2Info.setSoapNamespaceURI(messageContext.getEnvelope().getNamespace().getNamespaceURI());
        axis2Info.setAddressingNamespaceURI(addressingNamespaceURI);
        axis2Info.setProperties(messageContext.getProperties());
        return axis2Info;
    }

    public synchronized RMDSequence createRMDSequenceWithOffer(
            String offerID,
            String acksToAddress,
            Axis2Info axis2Info,
            boolean isAnonymous){
        // creates a sequence object for this sequence
        RMDSequence rmdSequence = new RMDSequence(RMDSequence.STATE_10);
        rmdSequence.setAcksTo(new EndpointReference(acksToAddress));
        rmdSequence.setSequenceID(offerID);
        rmdSequence.setAxis2Info(axis2Info);
        rmdSequence.setAnonymous(isAnonymous);
        // set the invoker buffer 
        rmdSequence.getInvokerBuffer().setAnonymous(isAnonymous);

        registerRMDSequenceToSequenceID(offerID,rmdSequence);

        if (!isAnonymous) {
            // start the RMDSequenceWorker to poll the seqence.
            // we have to start this only for duplex mode sequences.
            RMDSequenceWorker rmdSequenceWorker = new RMDSequenceWorker(rmdSequence);
            configurationContext.getThreadPool().execute(rmdSequenceWorker);
        }

        // start the InvokerWorker to pass the message to next level
        InvokerWorker invokerWorker = new InvokerWorker(rmdSequence.getInvokerBuffer());
        configurationContext.getThreadPool().execute(invokerWorker);

        return rmdSequence;
    }

    public synchronized void removeExpiredSequences(){

        List expiredSequences = new ArrayList();

        String sequenceID = null;
        RMDSequence rmdSequence = null;
        for (Iterator iter = this.sequenceIDRMDSequenceMap.keySet().iterator(); iter.hasNext();) {
            sequenceID = (String) iter.next();
            rmdSequence = (RMDSequence) this.sequenceIDRMDSequenceMap.get(sequenceID);
            if ((rmdSequence.getState() == RMDSequence.STATE_TERMINATED) ||
                ((System.currentTimeMillis() - rmdSequence.getLastAccesedTime()) > RMDSequence.TIMEOUT_TIME)) {
                // this sequence is expired
                expiredSequences.add(sequenceID);
            }
        }

        for (Iterator iter = expiredSequences.iterator();iter.hasNext();){
            sequenceID = (String) iter.next();
            log.info("Removing the sequence " + sequenceID + " from the RMDContext ");
            this.sequenceIDRMDSequenceMap.remove(sequenceID);
        }
    }

    public void registerRMDSequenceToSequenceID(String sequenceID, RMDSequence rmdSequence) {
        this.sequenceIDRMDSequenceMap.put(sequenceID, rmdSequence);
    }

    public synchronized RMDSequence getRMDSeqenceWithSequenceID(String sequenceID) {
        return (RMDSequence) this.sequenceIDRMDSequenceMap.get(sequenceID);
    }

    public void unRegisterRMDSequenceWithSequenceID(String sequenceID) {
        this.sequenceIDRMDSequenceMap.remove(sequenceID);
    }

    public ConfigurationContext getConfigurationContext() {
        return configurationContext;
    }

    public void setConfigurationContext(ConfigurationContext configurationContext) {
        this.configurationContext = configurationContext;
    }

}
