/*
 * Copyright 2008,2009 WSO2, Inc. http://www.wso2.org.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.mercury.state;

import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.OperationContext;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.util.MessageContextBuilder;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.transport.RequestResponseTransport;
import org.apache.axis2.wsdl.WSDLConstants;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axiom.om.util.UUIDGenerator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.mercury.context.MercuryMessageContext;
import org.wso2.mercury.message.*;
import org.wso2.mercury.util.MercuryConstants;
import org.wso2.mercury.workers.MessageWorker;
import org.wso2.mercury.exception.RMMessageBuildingException;
import org.wso2.mercury.persistence.PersistenceManager;
import org.wso2.mercury.persistence.exception.PersistenceException;
import org.wso2.mercury.persistence.dto.RMDSequenceDto;
import org.wso2.mercury.persistence.dto.SequenceReceivedNumberDto;
import org.wso2.mercury.persistence.dto.InvokerBufferDto;

import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;

/**
 * this class keeps the sates of the RMD
 * state of the RMD is basically determine by two factors for
 * this model
 * FMR - first message received.
 * LMR - last message received.
 * these two factors forms possible 4 states. but only
 * 3 can be really exists.
 * <p/>
 * This stat machine stat is changed by possible following events
 * AMR(SC) - Application message received and this completes the sequence.
 * this means we have alrady receive the last message and there is one
 * missing message to complete the sequence.
 * AMR(SNC) - Application message received and it does not complete the sequence.
 * here it can either be a without receiving last message or there may be missing messages
 * LMR(SC) - Last message receive which completes the sequence
 * LMR(SNC) - Last message receive but sequence is not completed.
 */

public class RMDSequence {

    private static Log log = LogFactory.getLog(RMDSequence.class);

    // states of the sequence
    public static final int STATE_00 = 0;
    public static final int STATE_10 = 1;
    public static final int STATE_11 = 2;
    public static final int STATE_COMPLETED = 3;
    public static final int STATE_TERMINATED = 4;

    public static long RETRANSMIT_TIME = 4000;
    public static long TIMEOUT_TIME = 60000;


    private String sequenceID;
    private int state;

    // acks to is send by the RMS when creating the sequence. when sending
    // responses we have to send them to here.
    private EndpointReference acksTo;

    private long lastCreateSequceResponseMessageSentTime = 0;
    private long lastAcknowledgmentSentTime = 0;
    private long lastAccesedTime = 0;

    // this is the initial createSequenceMessage context we received.
    // this is used when sending any createSequenceResponse
    // this should be theoratically correct.
    // TODO: think whether we need to update this upon receing new create seqence requests.
    private MessageContext createSequenceMessageContext;

    private Axis2Info axis2Info;

    private Set receivedMessageNumbers;
    private long lastMessageNumber = 0;

    private InvokerBuffer invokerBuffer;

    private RMDSequenceDto persistanceDto;

    private boolean isAnonymous;

    // keeps whether the RMS has offered a sequence or not.
    // this is usefull when sending create sequence Response where we have to send
    // and Accept header.
    private String selfAcksToEPR;

    private RMSSequence offeredRMSSequence;

    public RMDSequence(int state) {
        this.state = state;
        this.receivedMessageNumbers = new HashSet();
        this.lastAccesedTime = System.currentTimeMillis();
        // creates a invoker buffer for this RMD
        this.invokerBuffer = new InvokerBuffer(InvokerBuffer.STATE_000);
    }

    public void setInvokerBufferPersistanceManager(){
        this.invokerBuffer.setPersistanceManager(getPersistanceManager());
    }

    public synchronized void doActions() throws AxisFault, RMMessageBuildingException {
        switch (state) {
            case STATE_00 : {
                //i.e still either the first message or last message has not received
                // the only possible thing is to send the create sequene response.
                if ((lastCreateSequceResponseMessageSentTime == 0) ||
                        ((System.currentTimeMillis() - lastCreateSequceResponseMessageSentTime) > RETRANSMIT_TIME)) {
                    // send the create sequence response message
                    sendCreateSequenceResponseMessage();
                    lastCreateSequceResponseMessageSentTime = System.currentTimeMillis();
                }
                break;
            }
            case STATE_10 : {
                // i.e we have received the first message and but not completed.
                // the only thing we can do here is to send an acknowledgment.
                retransmitSequenceAcknowledgment();
                break;
            }
            case STATE_11 : {
                // i.e we have received the first message and but not completed.
                // the only thing we can do here is to send an acknowledgment.
                retransmitSequenceAcknowledgment();
                break;
            }

            case STATE_COMPLETED : {
                // i.e we have received all the required messages but still terminate
                // message has not received. so RMS may not have got acknowledgments for
                // all messages.
                retransmitSequenceAcknowledgment();
                break;
            }
        }
    }

    public void sendCreateSequenceResponseMessage() throws AxisFault, RMMessageBuildingException {

        // TODO: set the correct mercury and soap namesapce
        if (this.createSequenceMessageContext != null) {
            // the create Sequence Message Context is not null only for the sequences started with the
            // create sequence reqest
            // note that the persistence storage sequences start straight away with the
            // message received state. therefore the createSequence responses don't have to send

            CreateSequenceResponseMessage createSequenceResponseMessage = new CreateSequenceResponseMessage();
            createSequenceResponseMessage.setIdentifier(sequenceID);
            createSequenceResponseMessage.setSoapNamesapce(
                    this.createSequenceMessageContext.getEnvelope().getNamespace().getNamespaceURI());

            // set the accept header if a sequenc has occured
            if (this.selfAcksToEPR != null){
                Accept accept = new Accept();
                accept.setAcceptERP(this.selfAcksToEPR);
                createSequenceResponseMessage.setAccept(accept);
            }


            MessageContext messageContext =
                    MessageContextBuilder.createOutMessageContext(this.createSequenceMessageContext);
            messageContext.setProperty(MercuryConstants.PROCESS_RM_CONTROL_MESSAGE, Constants.VALUE_TRUE);
            messageContext.setEnvelope(createSequenceResponseMessage.toSOAPEnvelope());

            messageContext.setTo(this.acksTo);
            messageContext.getOptions().setAction(MercuryConstants.CREATE_SEQUENCE_RESPONSE_ACTION);

            //use message worker to send this message.
            sendMessage(messageContext, false);
        }
    }

    private void sendMessage(MessageContext messageContext, boolean isResume) {
        MessageWorker messageWorker = new MessageWorker(messageContext, isResume);
        // for annonymous sequence messages we have to use the same thread
        if (isAnonymous){
            messageWorker.run();
        } else {
            messageContext.getConfigurationContext().getThreadPool().execute(messageWorker);
        }

    }

    /**
     * this message should not be a last message
     *
     * @param messageNumber
     */
    public synchronized boolean applicationMessageReceived(long messageNumber,
                                                        MercuryMessageContext mercuryMessageContext)
            throws PersistenceException {
        // if this message is the last message then set it.
        this.lastAccesedTime = System.currentTimeMillis();
        int currentState = this.state;
        boolean newMessageReceived = false;
        if (!this.receivedMessageNumbers.contains(new Long(messageNumber))){
            this.receivedMessageNumbers.add(new Long(messageNumber));
            newMessageReceived = true;
        }

        switch (state) {
            case STATE_00 : {
                state = STATE_10;
                if (this.persistanceDto != null){
                    this.persistanceDto.setState(state);
                }
                break;
            }
            case STATE_10 : {
                // nothing to do
                break;
            }
            case STATE_11 : {
                if (isAllMessagesReceived()) {
                    state = STATE_COMPLETED;
                    if (this.persistanceDto != null) {
                        this.persistanceDto.setState(state);
                    }
                }
                break;
            }
        }
        // pass this message to invoker buffer
        try {
            this.invokerBuffer.applicatinMessageReceived(messageNumber, mercuryMessageContext, this.persistanceDto);
        } catch (PersistenceException e) {
            this.state = currentState;
            this.persistanceDto.setState(currentState);
            if (newMessageReceived){
                this.receivedMessageNumbers.remove(new Long(messageNumber));
            }
            log.error("Can not save the new message to persistence", e);
            throw new PersistenceException("Can not save the new message to persistence", e);
        }
        return newMessageReceived;
    }

    public synchronized boolean lastMessageReceived(long messageNumber, MercuryMessageContext mercuryMessageContext) throws PersistenceException {
        // if this message is the last message then set it.
        this.lastAccesedTime = System.currentTimeMillis();

        // here we have to keep track of changes to roll back them if we could
        // not persists the state.

        boolean isLastMessageSet = false;
        if (this.lastMessageNumber > 0){
            isLastMessageSet = true;
        }
        this.lastMessageNumber = messageNumber;
        if (this.persistanceDto != null){
            this.persistanceDto.setLastMessageNumber(this.lastMessageNumber);
        }
        int currentState = this.state;
        boolean newMessageReceived = false;
        if (!this.receivedMessageNumbers.contains(new Long(messageNumber))){
            newMessageReceived = true;
            this.receivedMessageNumbers.add(new Long(messageNumber));
        }

        switch (state) {
            case STATE_00 : {
                if (isAllMessagesReceived()) {
                    state = STATE_COMPLETED;
                } else {
                    state = STATE_11;
                }
                if(this.persistanceDto != null){
                    this.persistanceDto.setState(state);
                }
                break;
            }
            case STATE_10 : {
                if (isAllMessagesReceived()) {
                    state = STATE_COMPLETED;
                } else {
                    state = STATE_11;
                }
                if(this.persistanceDto != null){
                    this.persistanceDto.setState(state);
                }
                break;
            }
            case STATE_11 : {
                // nothing to do
                break;
            }
        }

        // pass this message to invoker buffer.
        try {
            this.invokerBuffer.lastMessageReceived(messageNumber, mercuryMessageContext, this.persistanceDto);
        } catch (PersistenceException e) {
            this.state = currentState;
            this.persistanceDto.setState(this.state);
            if (!isLastMessageSet){
               this.lastMessageNumber = 0;
               this.persistanceDto.setLastMessageNumber(0);
            }
            if (newMessageReceived){
               this.receivedMessageNumbers.remove(new Long(messageNumber));
            }
            log.error("can not save the receced last message", e);
            throw new PersistenceException("can not save the receced last message", e);
        }
        return newMessageReceived;
    }


    private boolean isAllMessagesReceived() {
        boolean isComplete = false;
        if (lastMessageNumber > 0) {
            isComplete = (this.receivedMessageNumbers.size() == lastMessageNumber);
        }
        return isComplete;
    }

    private void retransmitSequenceAcknowledgment()
            throws AxisFault, RMMessageBuildingException {
        if ((lastAcknowledgmentSentTime == 0) ||
                ((System.currentTimeMillis() - lastAcknowledgmentSentTime) > RETRANSMIT_TIME)) {
            sendSequenceAcknowledgementMessage(null);
            lastAcknowledgmentSentTime = System.currentTimeMillis();
        }
    }

    public void sendSequenceAcknowledgementMessage(MessageContext inboundMessageContext)
            throws AxisFault, RMMessageBuildingException {

        // there is no point in sending an empty sequece acknowledgment message.
        if (this.receivedMessageNumbers.size() > 0) {
            SequenceAcknowledgment sequenceAcknowledgment = new SequenceAcknowledgment(sequenceID);
            sequenceAcknowledgment.populateAcknowledgmentRanges(this.receivedMessageNumbers);
            sequenceAcknowledgment.setSoapNamesapce(this.getAxis2Info().getSoapNamespaceURI());

            SequenceAcknowledgmentMessage sequenceAcknowledgmentMessage =
                    new SequenceAcknowledgmentMessage(sequenceAcknowledgment);
            sequenceAcknowledgmentMessage.setSoapNamesapce(this.getAxis2Info().getSoapNamespaceURI());

            MessageContext messageContext =
                    getNewMessageContext(sequenceAcknowledgmentMessage, inboundMessageContext);


            messageContext.getOptions().setAction(MercuryConstants.SEQUENCE_ACKNOWLEDGMENT_ACTION);
            AxisService axisServce = this.axis2Info.getAxisService();
            AxisOperation inOnlyOperation =
                    axisServce.getOperationByAction(MercuryConstants.SEQUENCE_ACKNOWLEDGMENT_ACTION);

            OperationContext operationContext =
                    this.axis2Info.getServiceContext().createOperationContext(inOnlyOperation);
            inOnlyOperation.registerOperationContext(messageContext, operationContext);

            messageContext.setAxisMessage(inOnlyOperation.getMessage(WSDLConstants.MESSAGE_LABEL_IN_VALUE));
            // we have to use the same message formatter type
            messageContext.setProperty(Constants.Configuration.MESSAGE_TYPE,
                    this.axis2Info.getProperty(Constants.Configuration.MESSAGE_TYPE));
            //sending the message
            sendMessage(messageContext, false);
        }
    }

    public synchronized SequenceAcknowledgment getSequenceAcknowledgment(){
        SequenceAcknowledgment sequenceAcknowledgment = new SequenceAcknowledgment(this.sequenceID);
        sequenceAcknowledgment.populateAcknowledgmentRanges(this.receivedMessageNumbers);
        sequenceAcknowledgment.setSoapNamesapce(this.axis2Info.getSoapNamespaceURI());
        return sequenceAcknowledgment;
    }

    private MessageContext getNewMessageContext(
            SequenceAcknowledgmentMessage sequenceAcknowledgmentMessage,
            MessageContext inboundMessageContext) throws AxisFault, RMMessageBuildingException {
        MessageContext messageContext = new MessageContext();
        messageContext.setServiceContext(this.axis2Info.getServiceContext());
        messageContext.setEnvelope(sequenceAcknowledgmentMessage.toSOAPEnvelope());
        messageContext.setMessageID(UUIDGenerator.getUUID());

        // setting the options
        messageContext.setProperty(MercuryConstants.PROCESS_RM_CONTROL_MESSAGE, Constants.VALUE_TRUE);
        if (inboundMessageContext == null) {
            // i.e. inbound Message context means this is a duplex sequence.
            messageContext.setTransportOut(this.axis2Info.getTransportOut());
            messageContext.getOptions().setTo(this.acksTo);
        } else {
            // i.e we have to send the message using the backchanel of the incomming message
            messageContext.setTransportOut(inboundMessageContext.getTransportOut());
            messageContext.setProperty(MessageContext.TRANSPORT_OUT,
                    inboundMessageContext.getProperty(MessageContext.TRANSPORT_OUT));
            messageContext.setProperty(Constants.OUT_TRANSPORT_INFO,
                    inboundMessageContext.getProperty(Constants.OUT_TRANSPORT_INFO));
            // this is very important in writting the message using the same channel
            // once response input stream is written it updates this property
            messageContext.setProperty(RequestResponseTransport.TRANSPORT_CONTROL,
                    inboundMessageContext.getProperty(RequestResponseTransport.TRANSPORT_CONTROL));
        }

        return messageContext;
    }

    public synchronized void terminateSequenceMessageReceived() throws PersistenceException {
        this.lastAccesedTime = System.currentTimeMillis();
        int currentState = this.state;
        switch (state) {
            case STATE_00 : {
                state = STATE_TERMINATED;
                if (this.persistanceDto != null){
                    this.persistanceDto.setState(this.state);
                }
                break;
            }
            case STATE_10 : {
                state = STATE_TERMINATED;
                if (this.persistanceDto != null){
                    this.persistanceDto.setState(this.state);
                }
                break;
            }
            case STATE_11 : {
                state = STATE_TERMINATED;
                if (this.persistanceDto != null){
                    this.persistanceDto.setState(this.state);
                }
                break;
            }
            case STATE_COMPLETED : {
                state = STATE_TERMINATED;
                if (this.persistanceDto != null){
                    this.persistanceDto.setState(this.state);
                }
                break;
            }
        }
        try {
            if (this.persistanceDto != null){
                this.persistanceDto.setEndTime(System.currentTimeMillis());
            }
            this.invokerBuffer.terminateMessageReceived(this.persistanceDto);
        } catch (PersistenceException e) {
            this.state = currentState;
            this.persistanceDto.setState(currentState);
            log.error("Can not update the persistence state", e);
            throw new PersistenceException("Can not update the persistence state", e);
        }
    }

    public void save() throws PersistenceException {
        PersistenceManager persistenceManager = getPersistanceManager();
        if (persistenceManager != null){
            this.persistanceDto = new RMDSequenceDto();
            this.persistanceDto.setSequenceID(this.sequenceID);
            this.persistanceDto.setState(this.state);
            this.persistanceDto.setAcksTo(this.acksTo.getAddress());
            this.persistanceDto.setLastMessageNumber(this.lastMessageNumber);
            this.persistanceDto.setStartTime(System.currentTimeMillis());
            this.invokerBuffer.save(this.persistanceDto);
        }

    }

    public void loadRMDSequenceDetails(MessageContext messageContext)
            throws PersistenceException, AxisFault {

        // load the SequenceReceivedNumber details
        PersistenceManager persistenceManager = getPersistanceManager();
        List sequenceReceivedNumbers =
                persistenceManager.getSequenceReceivedNumbersWithRMDSequenceID(this.persistanceDto.getId());
        SequenceReceivedNumberDto sequenceReceivedNumberDto;
        for (Iterator iter = sequenceReceivedNumbers.iterator(); iter.hasNext();) {
            sequenceReceivedNumberDto = (SequenceReceivedNumberDto) iter.next();
            this.receivedMessageNumbers.add(new Long(sequenceReceivedNumberDto.getNumber()));
        }

        InvokerBufferDto invokerBufferDto =
                persistenceManager.getInvokerBufferWithRMDSequenceID(this.persistanceDto.getId());
        this.invokerBuffer.setState(invokerBufferDto.getState());
        this.invokerBuffer.setLastMessageNumber(invokerBufferDto.getLastMessage());
        this.invokerBuffer.setLastMessageSendToApplication(invokerBufferDto.getLastMessageToApplication());
        this.invokerBuffer.setPersistanceDto(invokerBufferDto);

        //load the invokerBuffer message details
        this.invokerBuffer.loadInvokerBufferDetails(messageContext);



    }

    private PersistenceManager getPersistanceManager(){
        PersistenceManager persistenceManager = null;
        if (this.axis2Info.getConfigurationContext().getProperty(
                MercuryConstants.RM_PERSISTANCE_MANAGER) != null){
            persistenceManager = (PersistenceManager)
                    this.axis2Info.getConfigurationContext().getProperty(MercuryConstants.RM_PERSISTANCE_MANAGER);
        }
        return persistenceManager;
    }

    /**
     * this method is called for an unexpected sequence termination
     */
    public synchronized void terminate(){
        this.state = STATE_TERMINATED;
        this.invokerBuffer.terminate();
    }

    public synchronized boolean isMessageReceived(long messageNumber){
        return this.receivedMessageNumbers.contains(new Long(messageNumber));
    }

    public synchronized int getState() {
        return state;
    }

    public void setState(int state) {
        this.state = state;
    }

    public synchronized long getLastAccesedTime() {
        return lastAccesedTime;
    }

    public void setLastAccesedTime(long lastAccesedTime) {
        this.lastAccesedTime = lastAccesedTime;
    }

    public EndpointReference getAcksTo() {
        return acksTo;
    }

    public void setAcksTo(EndpointReference acksTo) {
        this.acksTo = acksTo;
    }

    public String getSequenceID() {
        return sequenceID;
    }

    public void setSequenceID(String sequenceID) {
        this.sequenceID = sequenceID;
    }

    public MessageContext getCreateSequenceMessageContext() {
        return createSequenceMessageContext;
    }

    public void setCreateSequenceMessageContext(MessageContext createSequenceMessageContext) {
        this.createSequenceMessageContext = createSequenceMessageContext;
    }

    public InvokerBuffer getInvokerBuffer() {
        return invokerBuffer;
    }

    public void setInvokerBuffer(InvokerBuffer invokerBuffer) {
        this.invokerBuffer = invokerBuffer;
    }

    public Axis2Info getAxis2Info() {
        return axis2Info;
    }

    public void setAxis2Info(Axis2Info axis2Info) {
        this.axis2Info = axis2Info;
    }

    public long getLastMessageNumber() {
        return lastMessageNumber;
    }

    public void setLastMessageNumber(long lastMessageNumber) {
        this.lastMessageNumber = lastMessageNumber;
    }

    public RMDSequenceDto getPersistanceDto() {
        return persistanceDto;
    }

    public void setPersistanceDto(RMDSequenceDto persistanceDto) {
        this.persistanceDto = persistanceDto;
    }

    public boolean isAnonymous() {
        return isAnonymous;
    }

    public void setAnonymous(boolean anonymous) {
        isAnonymous = anonymous;
    }

    public RMSSequence getOfferedRMSSequence() {
        return offeredRMSSequence;
    }

    public void setOfferedRMSSequence(RMSSequence offeredRMSSequence) {
        this.offeredRMSSequence = offeredRMSSequence;
    }

    public String getSelfAcksToEPR() {
        return selfAcksToEPR;
    }

    public void setSelfAcksToEPR(String selfAcksToEPR) {
        this.selfAcksToEPR = selfAcksToEPR;
    }



}
