/*
 * Copyright 2008,2009 WSO2, Inc. http://www.wso2.org.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.mercury.state;

import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.OperationContext;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.WSDL2Constants;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.transport.RequestResponseTransport;
import org.apache.axis2.engine.AxisEngine;
import org.apache.axis2.client.Options;
import org.apache.axis2.client.ServiceClient;
import org.apache.axis2.wsdl.WSDLConstants;
import org.apache.axiom.om.util.UUIDGenerator;
import org.apache.axiom.om.util.StAXUtils;
import org.apache.axiom.om.impl.builder.StAXBuilder;
import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAP12Constants;
import org.apache.axiom.soap.impl.builder.StAXSOAPModelBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.mercury.message.*;
import org.wso2.mercury.context.MercuryMessageContext;
import org.wso2.mercury.util.MercuryConstants;
import org.wso2.mercury.exception.RMMessageBuildingException;
import org.wso2.mercury.exception.RMSequenceCreationException;
import org.wso2.mercury.exception.SoapProcessingFaultException;
import org.wso2.mercury.workers.MessageWorker;
import org.wso2.mercury.workers.ErrorCallbackWorker;
import org.wso2.mercury.persistence.dto.RMSSequenceDto;
import org.wso2.mercury.persistence.dto.RMSMessageDto;
import org.wso2.mercury.persistence.PersistenceManager;
import org.wso2.mercury.persistence.exception.PersistenceException;
import org.wso2.mercury.callback.MercuryErrorCallback;

import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamException;
import java.util.*;
import java.io.ByteArrayInputStream;

/**
 * this class is used to keep the RMS sequence. RMS Sequence is
 * depends on four parameters
 * SS - Sequence Started
 * MIB - Messages in the buffer
 * TMR - terminate message received
 * LMR - last message received
 * Here we assume either TMR or LMR occurs.
 * Collectively these four state variables forms 16 possible states
 * but only 7 states can realy exists.
 * <p/>
 * External events that can change this state machine are
 * CRR - Create Sequence Response Receive
 * LMR - Last message receive
 * TMR - Terminate message receive
 * ACKALL - Receive Acknowledgments for all messages.
 * <p/>
 * when those event occurs they change the sate accordingly
 * seqence processor always go through this state table and do
 * the appropriate tasks according to the state.
 */

public class RMSSequence {

    private static Log log = LogFactory.getLog(RMSSequence.class);

    public static long RETRANSMIT_TIME;
    public static long TIMEOUT_TIME;

    // possible states for this state machine
    // this state machine state is depends on SS,MIB,TMR,LMR and 0
    // and 1 represents absent and present of each state
    public static final int STATE_0100 = 0;
    public static final int STATE_0101 = 1;
    public static final int STATE_0110 = 2;
    public static final int STATE_1110 = 3;
    public static final int STATE_1000 = 4;
    public static final int STATE_1100 = 5;
    public static final int STATE_1101 = 6;
    public static final int STATE_TERMINATE = 7;

    private int state;

    // this is used to keep a set of application message context objects
    // to be send.
    private Map messageBuffer;

    private long messageNumber = 0;
    private long lastMessageNumber;

    //timing information
    private long lastCreateSequnceMessageSendTime = 0;
    private long lastAccessedTime;

    private EndpointReference endPointReference;
    private EndpointReference ackToEpr;

    // WSRM sequence ID for this sequence
    private String sequenceID;

    private Axis2Info axis2Info;

    private String sequenceOffer;

    private RMSSequenceDto persistanceDto;

    private boolean isAnnonymous;

    private RMDSequence offeredRMDSequence;

    private MercuryErrorCallback errorCallback;

    public RMSSequence(int state,
                       EndpointReference endPointReference) {
        this.state = state;
        this.endPointReference = endPointReference;
        this.messageBuffer = new HashMap();
        this.lastAccessedTime = System.currentTimeMillis();
    }

    public synchronized void addRMMessageContext(RMApplicationMessage message,
                                                 MessageContext messageContext) throws PersistenceException {
        this.lastAccessedTime = System.currentTimeMillis();
        long number = ++messageNumber;
        if (this.persistanceDto != null) {
            this.persistanceDto.setMessageNumber(messageNumber);
        }
        Sequence sequence = message.getSequence();
        sequence.setMessageNumber(number);
        if (sequence.isLastMessage()) {
            lastMessageNumber = number;
            if (this.persistanceDto != null) {
                this.persistanceDto.setLastMessageNumber(lastMessageNumber);
            }
        }
        PersistenceManager persistenceManager = getPersistanceManager();
        RMSMessageDto rmsMessageDto = null;
        if (persistenceManager != null) {
            // persists this message.
            rmsMessageDto = new RMSMessageDto();
            rmsMessageDto.setMessageNumber(sequence.getMessageNumber());
            rmsMessageDto.setLastMessage(sequence.isLastMessage());
            rmsMessageDto.setSoapEnvelpe(message.getOriginalMessage().toString());
            rmsMessageDto.setSend(false);
            rmsMessageDto.setRmsSequenceID(this.persistanceDto.getId());
            try {
                persistenceManager.save(rmsMessageDto, this.persistanceDto);
            } catch (PersistenceException e) {
                messageNumber--;
                this.persistanceDto.setMessageNumber(messageNumber);
                if (sequence.isLastMessage()) {
                    this.lastMessageNumber = 0;
                    this.persistanceDto.setLastMessageNumber(0);
                }
                log.error("Can not save the message with message number "
                        + sequence.getMessageNumber(), e);
                throw new PersistenceException("Can not save the message with message number "
                        + sequence.getMessageNumber(), e);
            }
        }

        // store this message as MercuryMessageContext.
        this.messageBuffer.put(new Long(number),
                new MercuryMessageContext(rmsMessageDto, message, messageContext));
        // notify all the threads waiting for this object message receice.
        // this is important for annonymous inout operations.
        this.notifyAll();
    }

    /**
     * sender worker invokes this method time to time and excecute the correct actions
     */
    public synchronized void doActions() throws AxisFault, RMMessageBuildingException {
        switch (state) {
            case STATE_0100 : {
                retransmitCreateSequenceMessage();
                break;
            }
            case STATE_0101 : {
                retransmitCreateSequenceMessage();
                break;
            }
            case STATE_0110 : {
                retransmitCreateSequenceMessage();
                break;
            }
            case STATE_1110 : {
                // i.e. it has got the create sequence response and knows there are
                // no more messages from client but still messages in the buffer.
                retransmitApplicationMessages();
                break;
            }
            case STATE_1000 : {
                // nothing to do since buffer is empty
                break;
            }
            case STATE_1100 : {
                retransmitApplicationMessages();
                break;
            }
            case STATE_1101 : {
                retransmitApplicationMessages();
                break;
            }
        }
    }

    private void retransmitCreateSequenceMessage() throws AxisFault, RMMessageBuildingException {
        // i.e still create sequence response message has not received. but there
        // is a message in the buffer.
        // the only thing we have to do here is to send another createSequence message
        // if it is time to send.
        if ((lastCreateSequnceMessageSendTime == 0) || // this means no message has send
                ((System.currentTimeMillis() - lastCreateSequnceMessageSendTime) > RETRANSMIT_TIME)) {
            //we have to send a create sequence message
            sendCreateSequenceMessage();
            lastCreateSequnceMessageSendTime = System.currentTimeMillis();
        }
    }

    private void sendCreateSequenceMessage() throws AxisFault, RMMessageBuildingException {

        // create the create sequence message
        //TODO: set the mercury and addressing namesapces correctly
        CreateSequenceMessage createSequenceMessage = new CreateSequenceMessage();
        createSequenceMessage.setAcksToAddress(this.ackToEpr.getAddress());
        createSequenceMessage.setSoapNamesapce(this.axis2Info.getSoapNamespaceURI());
        createSequenceMessage.setAddressingNamespace(this.axis2Info.getAddressingNamespaceURI());
        createSequenceMessage.setOfferIdentifier(this.sequenceOffer);

        MessageContext messageContext = getNewMessageContextUsingAxis2Info(createSequenceMessage);
        messageContext.getOptions().setAction(MercuryConstants.CREATE_SEQUENCE_ACTION);

        AxisOperation axisOperation = null;
        if (this.axis2Info.isServerSide()){
            axisOperation = this.axis2Info.getAxisService().getOperationByAction(
                    MercuryConstants.CREATE_SEQUENCE_ACTION);
        } else {
            axisOperation = this.axis2Info.getAxisService().getOperation(ServiceClient.ANON_OUT_IN_OP);
        }

        OperationContext operationContext =
                this.axis2Info.getServiceContext().createOperationContext(axisOperation);

        axisOperation.registerOperationContext(messageContext, operationContext);
        messageContext.setAxisMessage(axisOperation.getMessage(WSDLConstants.MESSAGE_LABEL_OUT_VALUE));

        //register this message with the RMSContext
        RMSContext rmsContext =
                (RMSContext) messageContext.getConfigurationContext().getProperty(MercuryConstants.RMS_CONTEXT);
        rmsContext.registerRMSSequenceToMessageID(messageContext.getMessageID(), this);

        //sending the message
        if (!this.axis2Info.isServerSide() && isAnnonymous) {
            MessageContext responseMessageContext = messageContext.getConfigurationContext().createMessageContext();
            responseMessageContext.setServerSide(false);
            axisOperation.registerOperationContext(responseMessageContext, operationContext);
            responseMessageContext.setOptions(messageContext.getOptions());
            responseMessageContext.getOptions().setAction(MercuryConstants.CREATE_SEQUENCE_RESPONSE_ACTION);
            responseMessageContext.setMessageID(UUIDGenerator.getUUID());
            responseMessageContext.setServiceContext(messageContext.getServiceContext());
            responseMessageContext.setAxisMessage(axisOperation.getMessage(WSDLConstants.MESSAGE_LABEL_IN_VALUE));
            sendMessage(messageContext, responseMessageContext, false, true, false);
        } else {
            sendMessage(messageContext, null, false, false, false);
        }


    }

    /**
     * create sequence message received event has occured do the necessary state changes accordingly.
     * this event is only relavant to the sequence started state.
     */
    public synchronized void createSequenceResponseReceived() throws PersistenceException {
        this.lastAccessedTime = System.currentTimeMillis();
        int currentState = state;
        switch (state) {
            case STATE_0100 : {
                // advance state to 1100
                state = STATE_1100;
                break;
            }
            case STATE_0101 : {
                state = STATE_1101;
                break;
            }
            case STATE_0110 : {
                state = STATE_1110;
                break;
            }
            case STATE_1110 : {
                break;
            }
            case STATE_1000 : {
                break;
            }
            case STATE_1100 : {
                break;
            }
            case STATE_1101 : {
                break;
            }
        }

        PersistenceManager persistenceManager = getPersistanceManager();
        if (persistenceManager != null) {
            this.persistanceDto.setState(this.state);
            this.persistanceDto.setLastAccessedTime(this.lastAccessedTime);
            this.persistanceDto.setSequenceID(this.sequenceID);
            try {
                persistenceManager.update(this.persistanceDto);
            } catch (PersistenceException e) {
                // roll back the state
                this.persistanceDto.setState(currentState);
                this.state = currentState;
                this.sequenceID = null;
                log.error("Can not updated the RMS state for received create sequene " +
                        "response message ", e);
                throw new PersistenceException("Can not updated the RMS state for received create sequene " +
                        "response message ", e);
            }
        }
    }

    private void sendMessage(MessageContext messageContext,
                             MessageContext responseMessageContext,
                             boolean isResume,
                             boolean isInvokeAsAnonClient,
                             boolean isUseSameThread) {
        MessageWorker messageWorker =
                new MessageWorker(messageContext, responseMessageContext, isResume, isInvokeAsAnonClient);
        if (isInvokeAsAnonClient){
            messageWorker.setRmsSequence(this);
        }
        if (isUseSameThread) {
            messageWorker.run();
        } else {
            // we need to invoke this as annonymous client
            messageContext.getConfigurationContext().getThreadPool().execute(messageWorker);
        }
    }

    private MessageContext getNewMessageContextUsingAxis2Info(CreateSequenceMessage createSequenceMessage)
            throws AxisFault, RMMessageBuildingException {
        MessageContext messageContext = new MessageContext();
        messageContext.setServiceContext(this.axis2Info.getServiceContext());
        messageContext.setEnvelope(createSequenceMessage.toSOAPEnvelope());
        messageContext.setMessageID(UUIDGenerator.getUUID());
        messageContext.setTransportIn(this.axis2Info.getTransportIn());
        messageContext.setTransportOut(this.axis2Info.getTransportOut());
        messageContext.setReplyTo(this.ackToEpr);
        messageContext.setProperty(MercuryConstants.PROCESS_RM_CONTROL_MESSAGE, Constants.VALUE_TRUE);
        messageContext.setOptions(new Options());
        // set the correct options
        messageContext.getOptions().setTo(this.axis2Info.getOptions().getTo());
        messageContext.getOptions().setUseSeparateListener(this.axis2Info.getOptions().isUseSeparateListener());
        messageContext.setServerSide(this.axis2Info.isServerSide());
        return messageContext;
    }

    public void removeRelationShips(Options options) {
        if (options.getParent() != null) {
            removeRelationShips(options.getParent());
        }
        options.setRelationships(null);
    }

    private void retransmitApplicationMessages() throws AxisFault, RMMessageBuildingException {
        MercuryMessageContext mercuryMessageContext = null;
        RMApplicationMessage message = null;
        MessageContext applicationMessageContext = null;
        for (Iterator iter = messageBuffer.values().iterator(); iter.hasNext();) {
            mercuryMessageContext = (MercuryMessageContext) iter.next();
            long lastMessageSentTime = mercuryMessageContext.getLastMessageSendTime();
            if ((lastMessageSentTime == 0) || // this means this has no transmitted
                    ((System.currentTimeMillis() - lastMessageSentTime) > RETRANSMIT_TIME)) {
                // have to retransmit the message
                message = mercuryMessageContext.getRmApplicationMessage();
                message.getSequence().setSequenceID(sequenceID);
                applicationMessageContext = mercuryMessageContext.getMessageContext();
                applicationMessageContext.setCurrentHandlerIndex(mercuryMessageContext.getCurrentHandlerIndex());

                // if this has an offered sequence we have to append the sequence
                // acknowledgment message
                if (this.offeredRMDSequence != null){
                    SequenceAcknowledgment sequenceAcknowledgment =
                            this.offeredRMDSequence.getSequenceAcknowledgment();
                    if (sequenceAcknowledgment.getAcknowledgmentRanges().size() > 0){
                        message.setSequenceAcknowledgment(sequenceAcknowledgment);
                    }
                }

                applicationMessageContext.setEnvelope(message.toSOAPEnvelope());
                applicationMessageContext.setProperty(
                        MercuryConstants.PROCESS_RM_CONTROL_MESSAGE, Constants.VALUE_TRUE);

                // when sending in only application messages in annonymous client
                // the returend message is a sequence acknowledgment.
                // have to set this before sending messages.
                // async messages waits util response comes.
                mercuryMessageContext.setLastMessageSendTime(System.currentTimeMillis());
                if (!this.axis2Info.isServerSide() && isAnnonymous) {
                    if (applicationMessageContext.getAxisOperation().getMessageExchangePattern().equals(
                            WSDL2Constants.MEP_URI_OUT_ONLY)) {
                        MessageContext responseMessageContext =
                                this.axis2Info.getConfigurationContext().createMessageContext();
                        responseMessageContext.setServerSide(false);
                        responseMessageContext.setServiceContext(this.axis2Info.getServiceContext());
                        responseMessageContext.setMessageID(UUIDGenerator.getUUID());

                        AxisService axisServce = this.axis2Info.getAxisService();
                        AxisOperation inOnlyOperation = null;
                        if (applicationMessageContext.getOptions().getAction().equals(MercuryConstants.LAST_MESSAGE_ACTION) &&
                                (this.offeredRMDSequence != null)) {
                             // this means this is a last message and in out message so we have to
                             // set the last message to get the message
                            responseMessageContext.getOptions().setAction(MercuryConstants.LAST_MESSAGE_ACTION);
                            inOnlyOperation = axisServce.getOperationByAction(MercuryConstants.LAST_MESSAGE_ACTION);
                        } else {
                            responseMessageContext.getOptions().setAction(MercuryConstants.SEQUENCE_ACKNOWLEDGMENT_ACTION);
                            inOnlyOperation = axisServce.getOperationByAction(MercuryConstants.SEQUENCE_ACKNOWLEDGMENT_ACTION);
                        }


                        OperationContext operationContext =
                                this.axis2Info.getServiceContext().createOperationContext(inOnlyOperation);
                        inOnlyOperation.registerOperationContext(responseMessageContext, operationContext);

                        responseMessageContext.setAxisMessage(
                                inOnlyOperation.getMessage(WSDLConstants.MESSAGE_LABEL_IN_VALUE));
                        // set this as the piggy back message for out going message context
                        // this is usful only to smtp transport
                        applicationMessageContext.setProperty(Constants.PIGGYBACK_MESSAGE, responseMessageContext);

                        sendMessage(applicationMessageContext, responseMessageContext, true, true, false);
                    } else {
                        // for inout operation it should be the response message context for the operation
                        MessageContext responseMessageContext =
                                applicationMessageContext.getOperationContext().getMessageContext(
                                        WSDLConstants.MESSAGE_LABEL_IN_VALUE);

                      sendMessage(applicationMessageContext, responseMessageContext, true, true, false);
                    }

                } else {
                    // this is a normal duplex channel mode inonly message
                    sendMessage(applicationMessageContext, null, true, false, false);
                }

            }

            try {
                Thread.sleep(5000);
            } catch (InterruptedException e) {}

        }
    }

    public synchronized void sendApplicationMessageResponse(MessageContext inboundMessageContext,
                                                            long inMessageNumber)
            throws AxisFault, RMMessageBuildingException {
       // here we have to keep in mind that invoker buffer always gurantee the inorder delivary. therefore
       // the first message always corresponds to the first message in the return sequence
        if (this.messageBuffer.containsKey(new Long(inMessageNumber))) {
            // if the message buffer has this message number then we can send this using
            // orginal message context
            MercuryMessageContext mercuryMessageContext = (MercuryMessageContext)
                    this.messageBuffer.get(new Long(inMessageNumber));
            RMApplicationMessage message = mercuryMessageContext.getRmApplicationMessage();
            message.getSequence().setSequenceID(sequenceID);
            MessageContext outBoundMessageContext = mercuryMessageContext.getMessageContext();
            outBoundMessageContext.setCurrentHandlerIndex(mercuryMessageContext.getCurrentHandlerIndex());

            // set the sequence acknowledgment
            if (this.offeredRMDSequence != null) {
                SequenceAcknowledgment sequenceAcknowledgment =
                        this.offeredRMDSequence.getSequenceAcknowledgment();
                if (sequenceAcknowledgment.getAcknowledgmentRanges().size() > 0){
                    message.setSequenceAcknowledgment(sequenceAcknowledgment);
                }
            }

            outBoundMessageContext.setEnvelope(message.toSOAPEnvelope());
            outBoundMessageContext.setProperty(
                    MercuryConstants.PROCESS_RM_CONTROL_MESSAGE, Constants.VALUE_TRUE);

            // copy the transport details.
            // i.e we have to send the message using the backchanel of the incomming message
            outBoundMessageContext.setTransportOut(inboundMessageContext.getTransportOut());
            outBoundMessageContext.setProperty(MessageContext.TRANSPORT_OUT,
                    inboundMessageContext.getProperty(MessageContext.TRANSPORT_OUT));
            outBoundMessageContext.setProperty(Constants.OUT_TRANSPORT_INFO,
                    inboundMessageContext.getProperty(Constants.OUT_TRANSPORT_INFO));
            // this is very important in writting the message using the same channel
            // once response input stream is written it updates this property
            outBoundMessageContext.setProperty(RequestResponseTransport.TRANSPORT_CONTROL,
                    inboundMessageContext.getProperty(RequestResponseTransport.TRANSPORT_CONTROL));
            try {
                AxisEngine.resumeSend(outBoundMessageContext);
            } catch (Exception e) {
                e.printStackTrace();
            }

            mercuryMessageContext.setLastMessageSendTime(System.currentTimeMillis());
        } else {
           if (inMessageNumber > messageNumber ){
               // this means that response message for this inMessage number has not yet received.
               // so wait until mesage receives
               try {
                   // this wait will notify in the last message receive and message received events
                   this.wait();
                   // then try again
                   sendApplicationMessageResponse(inboundMessageContext, inMessageNumber);
               } catch (InterruptedException e) {}
           } else {
               // this means an acknowledgment for this message is already received .
               // that means receiver has got the message correctly. so send the
               // sequence acknowledgment only.
               if (this.offeredRMDSequence != null){
                   this.offeredRMDSequence.sendSequenceAcknowledgementMessage(inboundMessageContext);
               }
           }
       }

    }

    public synchronized void sequenceAcknowledgmentReceived(List acknowledgmentRanges) throws PersistenceException {
        this.lastAccessedTime = System.currentTimeMillis();
        if (this.persistanceDto != null) {
            this.persistanceDto.setLastAccessedTime(this.lastAccessedTime);
        }
        // update the state of the RMSSequenceDto
        switch (state) {
            case STATE_0100 : {
                //TODO: we can not get an acknowledgment before starting the sequence
                log.error("Acknowledgment received before statring the sequence");
                break;
            }
            case STATE_0101 : {
                //TODO: we can not get an acknowledgment before starting the sequence
                log.error("Acknowledgment received before statring the sequence");
                break;
            }
            case STATE_0110 : {
                //TODO: we can not get an acknowledgment before starting the sequence
                log.error("Acknowledgment received before statring the sequence");
                break;
            }
            case STATE_1110 : {
                processRangesAndUpdateState(acknowledgmentRanges, STATE_TERMINATE);
                break;
            }
            case STATE_1000 : {
                // nothing todo
                break;
            }
            case STATE_1100 : {
                processRangesAndUpdateState(acknowledgmentRanges, STATE_1000);
                break;
            }
            case STATE_1101 : {
                processRangesAndUpdateState(acknowledgmentRanges, STATE_TERMINATE);
                break;
            }
        }
    }

    private void processRangesAndUpdateState(List acknowledgmentRanges, int newState)
            throws PersistenceException {
        Set acknowledgedMessages = processAcknowledgeMessages(acknowledgmentRanges);
        PersistenceManager persistenceManager = getPersistanceManager();
        if (persistenceManager != null) {
            // update the state in database
            if (this.messageBuffer.size() == acknowledgedMessages.size()) {
                // this means all the messaes has been acknoledged
                this.persistanceDto.setState(newState);
            }
            Set acknolwdgeMessageDtos = new HashSet();
            MercuryMessageContext mercuryMessageContext = null;
            for (Iterator iter = acknowledgedMessages.iterator(); iter.hasNext();) {
                mercuryMessageContext = (MercuryMessageContext) this.messageBuffer.get(iter.next());
                mercuryMessageContext.getRmsPersistanceDto().setSend(true);
                acknolwdgeMessageDtos.add(mercuryMessageContext.getRmsPersistanceDto());
            }
            // update the states
            try {
                persistenceManager.updateMessagesAsSend(acknolwdgeMessageDtos, this.persistanceDto);
            } catch (PersistenceException e) {
                this.persistanceDto.setState(this.state);
                for (Iterator iter = acknowledgedMessages.iterator(); iter.hasNext();) {
                    mercuryMessageContext = (MercuryMessageContext) this.messageBuffer.get(iter.next());
                    mercuryMessageContext.getRmsPersistanceDto().setSend(false);
                }
                log.error("Can not update the RMSMessages", e);
                throw new PersistenceException("Can not update the RMSMessages", e);
            }

        }

        // remove these messages from the internal buffer
        for (Iterator iter = acknowledgedMessages.iterator(); iter.hasNext();) {
            this.messageBuffer.remove(iter.next());
        }

        if (this.messageBuffer.size() == 0) {
            this.state = newState;
        }
    }

    private Set processAcknowledgeMessages(List acknowledgmentRanges) {
        // remove the buffered messages if they have acknowledged
        Set acknowledgedMessages = new HashSet();
        Long key;
        for (Iterator iter = this.messageBuffer.keySet().iterator(); iter.hasNext();) {
            key = (Long) iter.next();
            if (isNumberAcknowledged(key.longValue(), acknowledgmentRanges)) {
                acknowledgedMessages.add(key);
            }
        }
        return acknowledgedMessages;
    }

    public boolean isNumberAcknowledged(long number, List acknowledgments) {
        boolean isNumberAcknowledged = false;
        AcknowledgmentRange acknowledgmentRange = null;
        for (Iterator iter = acknowledgments.iterator(); iter.hasNext();) {
            acknowledgmentRange = (AcknowledgmentRange) iter.next();
            if (acknowledgmentRange.isNumberInRange(number)) {
                isNumberAcknowledged = true;
                break;
            }
        }
        return isNumberAcknowledged;
    }

    public synchronized void applicationMessageReceivedFromClient(
            RMApplicationMessage message,
            MessageContext messageContext) throws PersistenceException {
        this.lastAccessedTime = System.currentTimeMillis();
        if (this.persistanceDto != null) {
            this.persistanceDto.setLastAccessedTime(this.lastAccessedTime);
        }
        int currentState = state;
        try {
            switch (state) {
                case STATE_0100 : {
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_0101 : {
                    state = STATE_1101;
                    if (this.persistanceDto != null) {
                        this.persistanceDto.setState(state);
                    }
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_0110 : {
                    // it is not possible to get application messages now.
                    // TODO: throw exception
                    log.error("Receiving an application message after receiving the terminate message");
                    break;
                }
                case STATE_1110 : {
                    // it is not possible to get application messages now.
                    // TODO: throw exception
                    log.error("Receiving an application message after receiving the terminate message");
                    break;
                }
                case STATE_1000 : {
                    state = STATE_1100;
                    if (this.persistanceDto != null) {
                        this.persistanceDto.setState(state);
                    }
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_1100 : {
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_1101 : {
                    addRMMessageContext(message, messageContext);
                    break;
                }
            }
        } catch (PersistenceException e) {
            // restore the state
            this.state = currentState;
            this.persistanceDto.setState(currentState);
            log.error("Can not add the message ", e);
            throw new PersistenceException("Can not add the message ", e);
        }
    }

    public synchronized void lastMessageReceivedFromClient(
            RMApplicationMessage message,
            MessageContext messageContext) throws PersistenceException {
        this.lastAccessedTime = System.currentTimeMillis();
        if (this.persistanceDto != null) {
            this.persistanceDto.setLastAccessedTime(this.lastAccessedTime);
        }
        int currentState = state;
        try {
            switch (state) {
                case STATE_0100 : {
                    state = STATE_0101;
                    if (this.persistanceDto != null) {
                        this.persistanceDto.setState(this.state);
                    }
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_0101 : {
                    //Two last messages recevied this is wrong
                    //TODO: throw an exception
                    log.error("Receiving a last message after one last message receives");
                    break;
                }
                case STATE_0110 : {
                    //terminate message recevied this is wrong
                    //TODO: throw an exception
                    log.error("Receiving a last message after one last message receives");
                    break;
                }
                case STATE_1110 : {
                    //terminate message recevied this is wrong
                    // TODO: throw exception
                    log.error("Receiving a last message after one last message receives");
                    break;
                }
                case STATE_1000 : {
                    state = STATE_1101;
                    if (this.persistanceDto != null) {
                        this.persistanceDto.setState(this.state);
                    }
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_1100 : {
                    state = STATE_1101;
                    if (this.persistanceDto != null) {
                        this.persistanceDto.setState(this.state);
                    }
                    addRMMessageContext(message, messageContext);
                    break;
                }
                case STATE_1101 : {
                    //Two last messages recevied this is wrong
                    //TODO: throw an exception
                    break;
                }
            }
        } catch (PersistenceException e) {
            // restore the state
            this.state = currentState;
            this.persistanceDto.setState(currentState);
            log.error("Can not save the last message ", e);
            throw new PersistenceException("Can not save the last message ", e);
        }
    }

    public synchronized void terminateMessageReceivedFromClient() throws PersistenceException {
        this.lastAccessedTime = System.currentTimeMillis();
        if (this.persistanceDto != null) {
            this.persistanceDto.setLastAccessedTime(this.lastAccessedTime);
        }
        int currentState = this.state;
        switch (state) {
            case STATE_0100 : {
                state = STATE_0110;
                break;
            }
            case STATE_0101 : {
                // we assume not to happen this
                break;
            }
            case STATE_0110 : {
                //terminate message recevied this is wrong
                //TODO: throw an exception
                log.error("Receiving a termiante message after getting one terminate message");
                break;
            }
            case STATE_1110 : {
                //terminate message recevied this is wrong
                // TODO: throw exception
                log.error("Receiving a termiante message after getting one terminate message");
                break;
            }
            case STATE_1000 : {
                state = STATE_TERMINATE;
                break;
            }
            case STATE_1100 : {
                state = STATE_1110;
                break;
            }
            case STATE_1101 : {
                // we expect not to happen this state.
                break;
            }
        }

        PersistenceManager persistenceManager = getPersistanceManager();
        if (persistenceManager != null) {
            this.persistanceDto.setState(this.state);
            try {
                persistenceManager.update(this.persistanceDto);
            } catch (PersistenceException e) {
                //rever the state
                this.persistanceDto.setState(currentState);
                this.state = currentState;
                log.error("Can not update the state", e);
                throw new PersistenceException("Can not update the state", e);
            }
        }
    }

    public synchronized void sendTerminateSequenceMessage(MessageContext inboundMessageContext)
            throws AxisFault, RMMessageBuildingException {

        // we have to send the terminate message only if this sequence has established correctly.
        if (this.sequenceID == null){
            log.info("Sequence has been terminated due to an error");
            return;
        }

        // set the seuqence end time
        // if there is a problem with the persistence we don't care
        log.info("Sending the termainate message for the sequence " + this.sequenceID);
        PersistenceManager persistenceManager = getPersistanceManager();
        if (persistenceManager != null) {
            this.persistanceDto.setLastAccessedTime(System.currentTimeMillis());
            this.persistanceDto.setEndTime(System.currentTimeMillis());
            try {
                persistenceManager.update(this.persistanceDto);
            } catch (PersistenceException e) {
                log.error("Error in updating the sequence to terminate state. how ever sequene" +
                        " is terminated correctly ");
            }
        }
        TerminateSequenceMessage terminateSequenceMessage = new TerminateSequenceMessage(sequenceID);
        terminateSequenceMessage.setSoapNamesapce(this.axis2Info.getSoapNamespaceURI());

        MessageContext messageContext =
                getNewMessageContextUsingAxis2Info(terminateSequenceMessage, inboundMessageContext);

        messageContext.getOptions().setAction(MercuryConstants.TERMINATE_SEQUENCE_ACTION);

        // here we have to use OutOnly operation otherwise smtp transport sender
        // may register a call back for this.
        AxisOperation axisOperation = null;
        if (this.axis2Info.isServerSide()) {
            axisOperation = this.axis2Info.getAxisService().getOperationByAction(
                    MercuryConstants.TERMINATE_SEQUENCE_ACTION);
            messageContext.setAxisMessage(axisOperation.getMessage(WSDLConstants.MESSAGE_LABEL_IN_VALUE));
        } else {
            axisOperation = this.axis2Info.getAxisService().getOperation(ServiceClient.ANON_OUT_ONLY_OP);
            messageContext.setAxisMessage(axisOperation.getMessage(WSDLConstants.MESSAGE_LABEL_OUT_VALUE));
        }

        OperationContext operationContext =
                this.axis2Info.getServiceContext().createOperationContext(axisOperation);
        axisOperation.registerOperationContext(messageContext, operationContext);

        //sending the message
        if (inboundMessageContext == null){
            // if there is no inbound messagecontext we send this message
            // using a new thread. i.e at the client side

            // if this is an annonymous and inout operation then we have to register
            // a response message context
            if (this.offeredRMDSequence != null) {
                MessageContext responseMessageContext =
                        this.axis2Info.getConfigurationContext().createMessageContext();
                responseMessageContext.setServerSide(false);
                responseMessageContext.setServiceContext(this.axis2Info.getServiceContext());
                responseMessageContext.setMessageID(UUIDGenerator.getUUID());

                OperationContext responseOperationContext =
                        this.axis2Info.getServiceContext().createOperationContext(axisOperation);
                responseMessageContext.getOptions().setAction(MercuryConstants.TERMINATE_SEQUENCE_ACTION);
                //for incomming messages we can use the terminate operation
                axisOperation = axisOperation = this.axis2Info.getAxisService().getOperationByAction(
                    MercuryConstants.TERMINATE_SEQUENCE_ACTION);
                axisOperation.registerOperationContext(responseMessageContext, responseOperationContext);

                responseMessageContext.setAxisMessage(
                        axisOperation.getMessage(WSDLConstants.MESSAGE_LABEL_IN_VALUE));

                // set the piggy back message
                messageContext.setProperty(Constants.PIGGYBACK_MESSAGE, responseMessageContext);

                sendMessage(messageContext, responseMessageContext, false, true, false);
            } else {
                sendMessage(messageContext, null, false, false, false);
            }
        } else {
            // for annonymous messages we use the same thread.
            // this happens at the server side.
            sendMessage(messageContext, null, false, false, true);
        }

    }

    private MessageContext getNewMessageContextUsingAxis2Info(
            TerminateSequenceMessage terminateSequenceMessage, MessageContext inboundMessageContext)
            throws AxisFault, RMMessageBuildingException {
        MessageContext messageContext = new MessageContext();
        messageContext.setServiceContext(this.axis2Info.getServiceContext());
        if (this.offeredRMDSequence != null){
            // this means this is an Annonymous in out sequence and we have to append the
            // sequence acknowledgement for it
           SequenceAcknowledgment sequenceAcknowledgment = this.offeredRMDSequence.getSequenceAcknowledgment();
           if (sequenceAcknowledgment.getAcknowledgmentRanges().size() > 0){
               terminateSequenceMessage.setSequenceAcknowledgment(sequenceAcknowledgment);
           }
        }
        messageContext.setEnvelope(terminateSequenceMessage.toSOAPEnvelope());
        messageContext.setMessageID(UUIDGenerator.getUUID());

        messageContext.setProperty(MercuryConstants.PROCESS_RM_CONTROL_MESSAGE, Constants.VALUE_TRUE);
        messageContext.setOptions(new Options());
        messageContext.getOptions().setUseSeparateListener(this.axis2Info.getOptions().isUseSeparateListener());
        messageContext.setServerSide(this.axis2Info.isServerSide());

        if (inboundMessageContext == null) {
            // i.e. inbound Message context means this is a duplex sequence.
            messageContext.setTransportOut(this.axis2Info.getTransportOut());
            messageContext.setTo(this.axis2Info.getOptions().getTo());
        } else {
            // set some security related properties
            // TODO: read these from the module xml file
            messageContext.setProperty("RECV_RESULTS", inboundMessageContext.getProperty("RECV_RESULTS"));
            // i.e we have to send the message using the backchanel of the incomming message
            messageContext.setProperty(Constants.Configuration.MESSAGE_TYPE,
                    inboundMessageContext.getProperty(Constants.Configuration.MESSAGE_TYPE));
            messageContext.setTransportOut(inboundMessageContext.getTransportOut());
            messageContext.setProperty(MessageContext.TRANSPORT_OUT,
                    inboundMessageContext.getProperty(MessageContext.TRANSPORT_OUT));
            messageContext.setProperty(Constants.OUT_TRANSPORT_INFO,
                    inboundMessageContext.getProperty(Constants.OUT_TRANSPORT_INFO));
            // this is very important in writting the message using the same channel
            // once response input stream is written it updates this property
            messageContext.setProperty(RequestResponseTransport.TRANSPORT_CONTROL,
                    inboundMessageContext.getProperty(RequestResponseTransport.TRANSPORT_CONTROL));
        }
        return messageContext;
    }

    public void save(long internalKeyID) throws PersistenceException {
        RMSSequenceDto rmsSequenceDto = new RMSSequenceDto();
        rmsSequenceDto.setState(this.state);
        rmsSequenceDto.setMessageNumber(this.messageNumber);
        rmsSequenceDto.setLastMessageNumber(this.lastMessageNumber);
        rmsSequenceDto.setEndPointAddress(this.endPointReference.getAddress());
        rmsSequenceDto.setAckToEpr(this.ackToEpr.getAddress());
        rmsSequenceDto.setSequenceOffer(this.sequenceOffer);
        rmsSequenceDto.setInternalKeyID(internalKeyID);
        rmsSequenceDto.setStartTime(System.currentTimeMillis());
        getPersistanceManager().save(rmsSequenceDto);
        this.persistanceDto = rmsSequenceDto;
    }

    private PersistenceManager getPersistanceManager() {
        PersistenceManager persistenceManager = null;
        if (this.axis2Info.getConfigurationContext().getProperty(
                MercuryConstants.RM_PERSISTANCE_MANAGER) != null) {
            persistenceManager = (PersistenceManager)
                    this.axis2Info.getConfigurationContext().getProperty(MercuryConstants.RM_PERSISTANCE_MANAGER);
        }
        return persistenceManager;
    }

    public void populatePersistnaceData(MessageContext messageContext, RMSSequenceDto rmsSequenceDto)
            throws PersistenceException, AxisFault {
        // set the RMS Sequence object status
        this.persistanceDto = rmsSequenceDto;
        this.sequenceID = rmsSequenceDto.getSequenceID();
        this.messageNumber = rmsSequenceDto.getMessageNumber();
        this.lastMessageNumber = rmsSequenceDto.getLastMessageNumber();
        this.ackToEpr = new EndpointReference(rmsSequenceDto.getAckToEpr());
        this.sequenceOffer = rmsSequenceDto.getSequenceOffer();
        this.lastAccessedTime = System.currentTimeMillis();

        // populate the message buffer
        PersistenceManager persistenceManager = getPersistanceManager();
        // persistence manager should not be null
        List rmsMessageDtos =
                persistenceManager.getRMSMessagesWithRMSSequenceID(this.persistanceDto.getId());
        RMSMessageDto rmsMessageDto = null;
        for (Iterator iter = rmsMessageDtos.iterator(); iter.hasNext();) {
            rmsMessageDto = (RMSMessageDto) iter.next();
            // creating the message context
            MessageContext newMessageContext = new MessageContext();

            newMessageContext.setCurrentHandlerIndex(messageContext.getCurrentHandlerIndex());
            newMessageContext.setCurrentPhaseIndex(messageContext.getCurrentPhaseIndex());
            newMessageContext.setExecutionChain(messageContext.getExecutionChain());

            newMessageContext.setServiceContext(messageContext.getServiceContext());
            newMessageContext.setMessageID(UUIDGenerator.getUUID());
            newMessageContext.setTransportIn(messageContext.getTransportIn());
            newMessageContext.setTransportOut(messageContext.getTransportOut());

            newMessageContext.setTo(messageContext.getTo());
            newMessageContext.setReplyTo(messageContext.getReplyTo());

            newMessageContext.setOptions(new Options());
            newMessageContext.getOptions().setTo(messageContext.getTo());
            newMessageContext.getOptions().setUseSeparateListener(messageContext.getOptions().isUseSeparateListener());
            newMessageContext.getOptions().setAction(messageContext.getOptions().getAction());
            newMessageContext.setServerSide(messageContext.isServerSide());


            AxisOperation inOutOperation = messageContext.getAxisOperation();
            OperationContext operationContext =
                    messageContext.getServiceContext().createOperationContext(inOutOperation);
            inOutOperation.registerOperationContext(newMessageContext, operationContext);
            messageContext.setAxisMessage(inOutOperation.getMessage(WSDLConstants.MESSAGE_LABEL_OUT_VALUE));

            RMApplicationMessage rmApplicationMessage =
                    new RMApplicationMessage(getSoapEnvelope(rmsMessageDto.getSoapEnvelpe()));
            Sequence sequence = rmApplicationMessage.getSequence();
            if (sequence == null){
                sequence = new Sequence();
                rmApplicationMessage.setSequence(sequence);
            }
            sequence.setLastMessage(rmsMessageDto.isLastMessage());
            sequence.setMessageNumber(rmsMessageDto.getMessageNumber());

            MercuryMessageContext mercuryMessageContext =
                    new MercuryMessageContext(rmsMessageDto, rmApplicationMessage, newMessageContext);
            this.messageBuffer.put(new Long(rmsMessageDto.getMessageNumber()), mercuryMessageContext);
        }

    }

    private SOAPEnvelope getSoapEnvelope(String soapEnvelpe)
            throws PersistenceException {
        try {
            XMLStreamReader xmlReader =
                    StAXUtils.createXMLStreamReader(new ByteArrayInputStream(soapEnvelpe.getBytes()));
            StAXBuilder builder = new StAXSOAPModelBuilder(xmlReader);
            SOAPEnvelope soapEnvelope = (SOAPEnvelope) builder.getDocumentElement();
            soapEnvelope.build();
            String soapNamespace = soapEnvelope.getNamespace().getNamespaceURI();
            if (soapEnvelope.getHeader() == null) {
                SOAPFactory soapFactory = null;
                if (soapNamespace.equals(SOAP12Constants.SOAP_ENVELOPE_NAMESPACE_URI)) {
                    soapFactory = OMAbstractFactory.getSOAP12Factory();
                } else {
                    soapFactory = OMAbstractFactory.getSOAP11Factory();
                }
                soapFactory.createSOAPHeader(soapEnvelope);
            }
            return soapEnvelope;
        } catch (XMLStreamException e) {
            log.error("Problem with the stored message", e);
            throw new PersistenceException("Problem with the stored message", e);
        }
    }

    public synchronized void invalidCreateSequenceResponseReceived() throws AxisFault {

        terminateRMSSequence();

        if (this.errorCallback != null){
            ErrorCallbackWorker errorCallbackWorker = new ErrorCallbackWorker(this.errorCallback,
                    new RMSequenceCreationException("Invalid Sequence Response Received"));
            this.axis2Info.getConfigurationContext().getThreadPool().execute(errorCallbackWorker);
        }

    }

    private void terminateRMSSequence() throws AxisFault {
        // set the state as terminated
        this.state = STATE_TERMINATE;

        // notify all the waiting threads
        MercuryMessageContext mercuryMessageContext = null;
        MessageContext messageContext = null;
        for (Iterator iter = this.messageBuffer.values().iterator(); iter.hasNext();){
            mercuryMessageContext = (MercuryMessageContext) iter.next();
            messageContext = mercuryMessageContext.getMessageContext();
            if (messageContext.getAxisOperation().getMessageExchangePattern().equals(
                    WSDL2Constants.MEP_URI_OUT_IN)){
                MessageContext inMessageContext = messageContext.getOperationContext().getMessageContext(
                    WSDLConstants.MESSAGE_LABEL_IN_VALUE);
                synchronized(inMessageContext){
                    inMessageContext.notify();
                }
            }
        }

        // terminate offeredRMD sequence if it is there
        if (this.offeredRMDSequence != null){
            this.offeredRMDSequence.terminate();
        }
    }

    public synchronized void soapFaultOccured() throws AxisFault {
        terminateRMSSequence();

        if (this.errorCallback != null) {
            ErrorCallbackWorker errorCallbackWorker = new ErrorCallbackWorker(this.errorCallback,
                    new SoapProcessingFaultException("Problem with processing the soap message"));
            this.axis2Info.getConfigurationContext().getThreadPool().execute(errorCallbackWorker);
        }

    }

    public String getSequenceID() {
        return sequenceID;
    }

    public void setSequenceID(String sequenceID) {
        this.sequenceID = sequenceID;
    }

    public EndpointReference getAckToEpr() {
        return ackToEpr;
    }

    public void setAckToEpr(EndpointReference ackToEpr) {
        this.ackToEpr = ackToEpr;
    }

    public synchronized int getState() {
        return state;
    }

    public void setState(int state) {
        this.state = state;
    }

    public synchronized long getLastAccessedTime() {
        return lastAccessedTime;
    }

    public void setLastAccessedTime(long lastAccessedTime) {
        this.lastAccessedTime = lastAccessedTime;
    }

    public Axis2Info getAxis2Info() {
        return axis2Info;
    }

    public void setAxis2Info(Axis2Info axis2Info) {
        this.axis2Info = axis2Info;
    }

    public String getSequenceOffer() {
        return sequenceOffer;
    }

    public void setSequenceOffer(String sequenceOffer) {
        this.sequenceOffer = sequenceOffer;
    }

    public RMSSequenceDto getPersistanceDto() {
        return persistanceDto;
    }

    public void setPersistanceDto(RMSSequenceDto persistanceDto) {
        this.persistanceDto = persistanceDto;
    }

    public boolean isAnnonymous() {
        return isAnnonymous;
    }

    public void setAnnonymous(boolean annonymous) {
        isAnnonymous = annonymous;
    }

    public RMDSequence getOfferedRMDSequence() {
        return offeredRMDSequence;
    }

    public void setOfferedRMDSequence(RMDSequence offeredRMDSequence) {
        this.offeredRMDSequence = offeredRMDSequence;
    }

    public MercuryErrorCallback getErrorCallback() {
        return errorCallback;
    }

    public void setErrorCallback(MercuryErrorCallback errorCallback) {
        this.errorCallback = errorCallback;
    }


}
